/*
	dotNET FontManager library
    Copyright (C) 2015 SCAND Ltd, svg@scand.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    This license has exceptions: it can be used with OxyPlot and SvgNet on basis 
    on Apache license. For all other cases it shoulkd be provided under GPLv3.
*/
#include <stdio.h>
#include <stdlib.h>
#include <ft2build.h>
#include <freetype/freetype.h>
#include <freetype/ftimage.h>
#include <freetype/ftbbox.h>
#include <freetype/ftoutln.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "conv.h"

class strbuf {
	char* buf;
	int capacity;
	int len;
	char* ptr;

public:
	static const int GAP = 1024;

	void reset() {
		if(buf) delete[] buf;
		buf = ptr = 0;
		len = capacity = 0;
	}

	strbuf() {
		buf = ptr = 0;
		len = capacity = 0;
	}

	~strbuf() {
		reset();
	}

	void append(char c) {
		if(len + 1 >= capacity) {
			char a[2];
			a[0] = c;
			a[1] = '\0';
			append(a);
		}
		else {
			*ptr++ = c;
            ++len;
		}
	}

	void append(char* s) {
		if(!s) return;
		int l = strlen(s);
		if(l <=0) return;
		if(len + l >= capacity) {
			capacity = len + l + GAP;
			char* v = new char[capacity];
			if(buf) {
				memset(v + len, 0, capacity - len);
				memcpy(v, buf, len);
				delete[] buf;
				buf = v;
				ptr = buf + len;
			}
			else {
				buf = ptr = v;
				len = 0;
			}
		}
		len += l;
		memcpy(ptr, s, l);
		ptr += l;
	}

	const char* copy() {
		char* v = new char[len + 1]; 
		memcpy(v, buf, len);
		v[len] = '\0';
		return (const char*)v;
	}

	int length() const { return len; }
};

class FtWalk {
	bool open;

	static const char C_END_PATH = 's'; 
	static const char C_MOVETO = 'M';
	static const char C_LINETO = 'L';
	static const char C_QUADTO = 'Q';
	static const char C_CURVETO = 'C';
	static const char C_START = 'G';
	static const char C_EOFILL = 'F';
	static const char C_BBOX ='B';
	static const char C_END_GLYPH = 'g';
	static const char C_EOL = '\n';
	static const char C_STARTFONT = 'f';
	static const char C_STARTMETRICS = 'x';

public:
	strbuf* str;
	FT_BBox box;

public:
	FtWalk() {
		str = new strbuf();
		open = false;
	}

	~FtWalk() {
		if(str) delete str;
		str = 0;
		open = false;
	}

	void end() {
		str->append(C_EOL);
	}

	void moveto(const FT_Vector* to) {
		if(open){ 
			str->append(C_END_PATH);
			end();
		}	

		str->append(C_MOVETO);
		v2s(to);
		open = true;
		end();
	}

	void lineto(const FT_Vector* to) {
		str->append(C_LINETO);
		v2s(to);
		end();
	}

	void quadto(const FT_Vector* t1, const FT_Vector* t2) {
		str->append(C_QUADTO);
		v2s(t1);
		v2s(t2);
		end();
	}

	void curveto(const FT_Vector* t1, const FT_Vector* t2, const FT_Vector* t3) {
		str->append(C_CURVETO);
		v2s(t1);
		v2s(t2);
		v2s(t3);
		end();
	}

	void start_glyph(int code, int uucode, FT_Vector* adv, int flags) {
		if(!str) str = new strbuf();

		str->append(C_START);
		i2s(code);
		v2s(adv);
		i2s(uucode);
		end();

		if((flags&FT_OUTLINE_EVEN_ODD_FILL) == FT_OUTLINE_EVEN_ODD_FILL) {
			str->append(C_EOFILL);
			end();
		}

		str->append(C_BBOX);
		f2s(box.xMin);
		f2s(box.yMin);
		f2s(box.xMax);
		f2s(box.yMax);
		end();
	}

	void end_glyph() {
		str->append(C_END_PATH);
		end();
		str->append(C_END_GLYPH);
		end();
		end();
		end();
	}

	void globalmetrics(FT_Short ascender, FT_Short descender, FT_Short height, 
		FT_Short units_per_EM, FT_Short max_advance_height) 
	{
		str->append(C_STARTFONT);
		i2s(ascender);
		i2s(descender);
		i2s(height);
		i2s(units_per_EM);
		i2s(max_advance_height);
		end();
		end();
		end();
	}

	void metrics(FT_Glyph_Metrics* fm) {
		str->append(C_STARTMETRICS);
		f2s(fm->width);
		f2s(fm->height);
		f2s(fm->horiBearingX);
		f2s(fm->horiAdvance);
		f2s(fm->horiBearingY);
		end();
	}

private:
	void i2s(int x) {
		char buf[32];
		itoa(x, buf, 10);
		str->append(' ');
		str->append(buf);
	}

	void f2s(FT_Pos x) {
		char buf[32];
		float y = (float)x /* / 64.0f */; // 64 = 6 bits
		str->append(' ');
		sprintf(buf, "%f", y);
		str->append(buf);
	}

	void v2s(const FT_Vector* to) {
		f2s(to->x);
		f2s(to->y);
	}
};

static int my_moveto(const FT_Vector* to, void* walk) {
	((FtWalk*)walk)->moveto(to);
	return 0;
}

static int my_lineto(const FT_Vector* to, void* walk) {
	((FtWalk*)walk)->lineto(to);
	return 0;
}

static int my_quadto(const FT_Vector* control,	const FT_Vector* to, void* walk) {
	((FtWalk*)walk)->quadto(control, to);
	return 0;
}

static int my_curveto(const FT_Vector* a,
		const FT_Vector* b,
		const FT_Vector* c,
		void* walk)
{
	((FtWalk*)walk)->curveto(a, b, c);
	return 0;
}

static FT_Outline_Funcs funcs = {
	my_moveto, my_lineto, my_quadto, my_curveto, 0, 0
};

const int CHAR_WIDTH_DEFAULT_SIZE  = 1000;
const int CHAR_HEIGHT_DEFAULT_SIZE = 1000;

const char* parseFont(void* bytes, int len)
{
	FT_Library fl;
	FT_Face face;
	FT_Open_Args fa;
	int cid_font = 0;
	int nglyphs = 0;
	int j, k;
	FT_Error ec;
	FtWalk walk;

	ec = FT_Init_FreeType(&fl);
	if(ec) return NULL;

	fa.flags = FT_OPEN_MEMORY; 
	fa.memory_base = (const FT_Byte*)bytes;
	fa.memory_size = len;
	fa.pathname = 0;
	fa.stream = 0;
	fa.driver = 0;
	fa.num_params = 0;
	fa.params = 0;

	ec = FT_Open_Face(fl, &fa, 0, &face);
	if(ec){ 
		FT_Done_FreeType(fl);
		return NULL;
	}

	cid_font = face->face_flags & FT_FACE_FLAG_CID_KEYED;
	nglyphs = face->num_glyphs;
	
	/* VK: bad idea for truetype fonts
	face->units_per_EM = 1;
	*/

	if( FT_IS_SCALABLE(face) ) {
        FT_Set_Char_Size( face, CHAR_WIDTH_DEFAULT_SIZE << 6, CHAR_HEIGHT_DEFAULT_SIZE << 6, 0, 0 );
        FT_Set_Transform( face, 0, 0 );
    }

    walk.globalmetrics(face->ascender, face->descender, face->height, face->units_per_EM,
    	face->max_advance_height);

    int* glySet = new int[nglyphs];
    for(j=0; j < nglyphs; ++j) glySet[j] = -1;
    
	FT_ULong cc = 0;
	FT_UInt  gindex = 0;
    int rgly = 0;

	cc = FT_Get_First_Char( face, &gindex );
	while ( gindex != 0 ) {
		int gly = FT_Get_Char_Index(face, cc);

		for(j=0; j < nglyphs; ++j) if(glySet[j] == gly) break;
		if(j >= nglyphs) 
			for(j=0; j < nglyphs; ++j) if(glySet[j] == -1){ 
				glySet[j] = gly;
				break;
			}
		
		// now we can load glyphs and walk with
		ec = FT_Load_Glyph(face, gly, FT_LOAD_NO_BITMAP|FT_LOAD_NO_AUTOHINT|FT_LOAD_NO_SCALE);
		if(ec == FT_Err_Invalid_Argument || ec){ 
			fprintf(stdout, "FT2: char `%c`: no glyph id %d\n", cc, gly);
			continue;
		}

		if((face->glyph->format&FT_GLYPH_FORMAT_OUTLINE) != FT_GLYPH_FORMAT_OUTLINE) {
			fprintf(stdout, "FT2: char `%c`: glyph id %d has no contours\n", cc, gly);
			continue;
		}

		ec = FT_Outline_Get_BBox(&face->glyph->outline, &walk.box);
		
		walk.start_glyph(j, cc, &face->glyph->advance, face->glyph->outline.flags);
		walk.metrics(&face->glyph->metrics);
		ec = FT_Outline_Decompose(&face->glyph->outline, &funcs, &walk);
		walk.end_glyph();

        ++rgly;

		// for
		cc = FT_Get_Next_Char( face, cc, &gindex );
	}
  
	for(j = 0; j < nglyphs; ++j) {
	    for(k=0; k < nglyphs; ++k) 
	    	if(glySet[k] == j) break;
	    if(k < nglyphs) continue;
	    for(k=0; k < nglyphs; ++k) 
	    	if(glySet[k] == -1) {
	    		glySet[k] = j;
	    		break;
	    	}

		ec = FT_Load_Glyph(face, j, FT_LOAD_NO_BITMAP|FT_LOAD_NO_AUTOHINT|FT_LOAD_NO_SCALE);
		if(ec == FT_Err_Invalid_Argument) continue; // normal case
		if(ec) continue; // such problem with font -> ignore glyph

		if((face->glyph->format&FT_GLYPH_FORMAT_OUTLINE) != FT_GLYPH_FORMAT_OUTLINE) continue; // not contours

		ec = FT_Outline_Get_BBox(&face->glyph->outline, &walk.box);
		
		walk.start_glyph(j, -j, &face->glyph->advance, face->glyph->outline.flags);
		walk.metrics(&face->glyph->metrics);
		ec = FT_Outline_Decompose(&face->glyph->outline, &funcs, &walk);
		walk.end_glyph();

        ++rgly;
	}
	delete[] glySet;

	FT_Done_Face(face);
	FT_Done_FreeType(fl);

	return walk.str->copy();
}
